/*
 *    sfall
 *    Copyright (C) 2008, 2009, 2010  The sfall team
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "main.h"

#include "HookScripts.h"
#include "vector9x.cpp"
#include "Logging.h"

#define MAXDEPTH (8)
static const int numHooks=17;

static vector<DWORD> hooks[numHooks];

DWORD InitingHookScripts;

static DWORD args[8];
static DWORD oldargs[8*MAXDEPTH];
static DWORD* argPtr;
static DWORD rets[8];

static DWORD firstArg=0;
static DWORD callDepth;
static DWORD lastCount[MAXDEPTH];

static DWORD ArgCount;
static DWORD cArg;
static DWORD cRet;

#define hookbegin(a) __asm pushad __asm call BeginHook __asm popad __asm mov ArgCount, a
#define hookend __asm pushad __asm call EndHook __asm popad

static void _stdcall BeginHook() {
	if(callDepth <= MAXDEPTH) {
		if(callDepth) {
			lastCount[callDepth-1]=ArgCount;
			memcpy(&oldargs[8*(callDepth-1)], args, 8*sizeof(DWORD));
		}
		argPtr=args;
		for(DWORD i=0;i<callDepth;i++) argPtr+=lastCount[i];
	}
	callDepth++;
}

static void _stdcall EndHook() {
	callDepth--;
	if(callDepth && callDepth <= MAXDEPTH) {
		ArgCount=lastCount[callDepth-1];
		memcpy(args, &oldargs[8*(callDepth-1)], 8*sizeof(DWORD));
	}
}

static void _stdcall RunSpecificHookScript(DWORD script) {
	cArg=0;
	cRet=0;
	__asm {
		mov edx, 1;
		mov eax, script;
		mov ebx, 0x46DD2C;
		call ebx;
	}
}
static void _stdcall RunHookScript(DWORD hook) {
	if(hooks[hook].size()) {
#ifdef TRACE
	char buf[256];
	sprintf_s(buf, "Running hook %d, which has %0d entries attached", hook, hooks[hook].size());
	dlogr(buf, DL_HOOK);
#endif
		for(int i=hooks[hook].size()-1;i>=0;i--) RunSpecificHookScript(hooks[hook][i]);
	} else {
		cArg=0;
		cRet=0;
	}
}

static const DWORD ToHitAddr=0x4243A8;
static void __declspec(naked) ToHitHook() {
	__asm {
		hookbegin(4);
		mov args[4], eax;
		mov args[8], ebx;
		mov args[12], ecx;
		push [esp+8];
		push [esp+8];
		call ToHitAddr;
		mov args[0], eax;
		pushad;
		push 0;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
end:
		hookend;
		retn 8;
	}
}
static const DWORD AfterHitRollAddr=0x423898;
static void __declspec(naked) AfterHitRollHook() {
	__asm {
		hookbegin(5);
		mov args[0], eax; //was it a hit?
		mov ebx, [esi];
		mov args[4], ebx; //Attacker
		mov ebx, [esi+0x20];
		mov args[8], ebx; //Target
		mov ebx, [esi+0x28];
		mov args[12], ebx; //bodypart
		mov ebx, [esp+0x18];
		mov args[16], ebx; //hit chance
		pushad;
		push 1;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
		cmp cRet, 2;
		jl end;
		mov ebx, rets[4];
		mov [esi+0x28], ebx;
		cmp cRet, 3;
		jl end;
		mov ebx, rets[8];
		mov [esi+0x20], ebx;
end:
		mov ebx, eax;
		hookend;
		cmp ebx, 1;
		jmp AfterHitRollAddr;
	}
}

static const DWORD CalcApCostAddr=0x478B24;
static void __declspec(naked) CalcApCostHook() {
	__asm {
		hookbegin(4);
		mov args[0], eax;
		mov args[4], edx;
		mov args[8], ebx;
		call CalcApCostAddr;
		mov args[12], eax;
		pushad;
		push 2;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
end:
		hookend;
		retn;
	}
}

static const DWORD ObjPidNewAddr=0x489C9C;
static const DWORD ObjEraseObjectAddr=0x48B0FC;
static const DWORD DeathAnimAddr=0x41060C;
static void __declspec(naked) CalcDeathAnimHook() {
	__asm {
		hookbegin(4);
		mov args[24], ebx;
		test ebx, ebx;
		jz noweap
		mov ebx, [ebx+0x64];
		and ebx, 0xfff;
		jmp weapend;
noweap:
		dec ebx;
weapend:
		mov args[0], ebx;
		mov ebx, args[24];
		mov args[4], eax;
		mov args[8], edx;
		mov args[12], ecx;
		mov args[20], 0;
		pushad;
		push 3;
		call RunHookScript;
		cmp cRet, 1;
		jl end1;
		sub esp, 4;
		mov edx, rets[0];
		mov args[0], edx;
		mov eax, esp;
		call ObjPidNewAddr;
		add esp, 4;
		cmp eax, 0xffffffff;
		jz end1;
		mov eax, [esp-4];
		mov args[20], 1;
		mov args[24], eax;
end1:
		popad;
		mov eax, [esp+8];
		mov ebx, [esp+4];
		push eax;
		push ebx;
		mov eax, args[4];
		mov ebx, args[24];
		call DeathAnimAddr;
		mov args[16], eax;
		mov eax, args[16];
		mov ArgCount, 5;
		pushad;
		push 4;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl skip2;
		mov eax, rets[0];
		mov args[16], eax;
skip2:
		mov eax, args[16];
		push eax;
		mov eax, args[20];
		test eax, eax;
		jz aend;
		mov eax, args[24];
		xor edx, edx;
		call ObjEraseObjectAddr;
aend:
		pop eax;
		hookend;
		retn 8;
	}
}
static const DWORD CombatDamageAddr=0x4247B8;
static void __declspec(naked) CombatDamageHook() {
	__asm {
		push eax;
		call CombatDamageAddr;
		pop edx;
		
		//zero damage insta death criticals fix
		mov ebx, [edx+0x2c];
		test ebx, ebx;
		jnz hookscript;
		mov ebx, [edx+0x30];
		test bl, 0x80;
		jz hookscript;
		inc dword ptr ds:[edx+0x2c];
hookscript:
		hookbegin(8);
		mov ebx, [edx+0x20];
		mov args[0x00], ebx;
		mov ebx, [edx+0x00];
		mov args[0x04], ebx;
		mov ebx, [edx+0x2c];
		mov args[0x08], ebx;
		mov ebx, [edx+0x10];
		mov args[0x0c], ebx;
		mov ebx, [edx+0x30];
		mov args[0x10], ebx;
		mov ebx, [edx+0x14];
		mov args[0x14], ebx;
		mov ebx, [edx+0x08];
		mov args[0x18], ebx;
		mov ebx, [edx+0x28];
		mov args[0x1c], ebx;

		pushad;
		push 5;
		call RunHookScript;
		popad;

		cmp cRet, 1;
		jl end;
		mov ebx, rets[0x00];
		mov [edx+0x2c], ebx;
		cmp cRet, 2;
		jl end;
		mov ebx, rets[0x04];
		mov [edx+0x10], ebx;
		cmp cRet, 3;
		jl end;
		mov ebx, rets[0x08];
		mov [edx+0x30], ebx;
		cmp cRet, 4;
		jl end;
		mov ebx, rets[0x0c];
		mov [edx+0x14], ebx;
end:
		hookend;
		retn;
	}
}
static const DWORD critter_kill=0x42DA64;
static void __declspec(naked) OnDeathHook() {
	__asm {
		hookbegin(1);
		mov args[0], eax;
		call critter_kill;
		pushad;
		push 6;
		call RunHookScript;
		popad;
		hookend;
		retn;
	}
}
static const DWORD OnDeathHook2Ret=0x4944DC;
static void __declspec(naked) OnDeathHook2() {
	__asm {
		hookbegin(1);
		mov args[0], esi;
		call OnDeathHook2Ret;
		pushad;
		push 6;
		call RunHookScript;
		popad;
		hookend;
		retn;
	}
}

static const DWORD _qsort=0x4F05B6;
static void __declspec(naked) FindTargetHook() {
	__asm {
		hookbegin(5);
		mov args[0], esi; //attacker
		mov edi, [eax+0];
		mov args[4], edi;
		mov edi, [eax+4];
		mov args[8], edi;
		mov edi, [eax+8];
		mov args[12], edi;
		mov edi, [eax+12];
		mov args[16], edi;
		pushad;
		push 7
		call RunHookScript;
		popad;
		cmp cRet, 4;
		jge cont;
		call _qsort;
		jmp end;
cont:
		mov edi, rets[0];
		mov [eax+0], edi;
		mov edi, rets[4];
		mov [eax+4], edi;
		mov edi, rets[8];
		mov [eax+8], edi;
		mov edi, rets[12];
		mov [eax+12], edi;
end:
		hookend;
		retn;
	}
}
static const DWORD _protinst_use_item_on=0x49C3CC;
static void __declspec(naked) UseObjOnHook() {
	__asm {
		hookbegin(3);
		mov args[0], edx;
		mov args[4], eax;
		mov args[8], ebx;
		pushad;
		push 8;
		call RunHookScript;
		popad;
		call _protinst_use_item_on;
		hookend;
		retn;
	}
}
static const DWORD RemoveObjHookRet=0x477497;
static void __declspec(naked) RemoveObjHook() {
	__asm {
		push ecx;
		mov ecx, [esp+4];
		hookbegin(4);
		mov args[0], eax;
		mov args[4], edx;
		mov args[8], ebx;
		mov args[12], ecx;
		pushad;
		push 9;
		call RunHookScript;
		popad;
		hookend;
		push esi;
		push edi;
		push ebp;
		sub esp, 0xc;
		jmp RemoveObjHookRet;
	}
}
static const DWORD _barter_compute_value=0x474B2C;
static const DWORD _item_caps_total=0x47A6A8;
static const DWORD _item_total_cost=0x477DAC;
static void __declspec(naked) BarterPriceHook() {
	__asm {
		hookbegin(6);
		mov args[0], eax;
		mov args[4], edx;
		call _barter_compute_value;
		mov edx, ds:[0x59E944];
		mov args[8], eax;
		mov args[12], edx;
		xchg eax, edx;
		call _item_caps_total;
		mov args[16], eax;
		mov eax, ds:[0x59E944];
		call _item_total_cost;
		mov args[20], eax;
		mov eax, edx;
		pushad;
		push 10;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
end:
		hookend;
		retn;
	}
}
static const DWORD _critter_compute_ap_from_distance=0x42E62C;
static void __declspec(naked) MoveCostHook() {
	__asm {
		hookbegin(3);
		mov args[0], eax;
		mov args[4], edx;
		call _critter_compute_ap_from_distance;
		mov args[8], eax;
		pushad;
		push 11;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
end:
		hookend;
		retn;
	}
}
static const DWORD _obj_blocking_at=0x48B84E;
static void __declspec(naked) HexMBlockingHook() {
	__asm {
		hookbegin(4);
		mov args[0], eax;
		mov args[4], edx;
		mov args[8], ebx;
		push next;
		push ecx;
		push esi;
		push edi;
		push ebp;
		mov ecx, eax;
		jmp _obj_blocking_at;
next:
		mov args[12], eax;
		pushad;
		push 12;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
end:
		hookend;
		retn;
	}
}
static const DWORD _obj_ai_blocking_at=0x48BA20;
static void __declspec(naked) HexABlockingHook() {
	__asm {
		hookbegin(4);
		mov args[0], eax;
		mov args[4], edx;
		mov args[8], ebx;
		call _obj_ai_blocking_at;
		mov args[12], eax;
		pushad;
		push 13;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
end:
		hookend;
		retn;
	}
}
static const DWORD _obj_shoot_blocking_at=0x48B930;
static void __declspec(naked) HexShootBlockingHook() {
	__asm {
		hookbegin(4);
		mov args[0], eax;
		mov args[4], edx;
		mov args[8], ebx;
		call _obj_shoot_blocking_at;
		mov args[12], eax;
		pushad;
		push 14;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
end:
		hookend;
		retn;
	}
}
static const DWORD _obj_sight_blocking_at=0x48BB88;
static void __declspec(naked) HexSightBlockingHook() {
	__asm {
		hookbegin(4);
		mov args[0], eax;
		mov args[4], edx;
		mov args[8], ebx;
		call _obj_sight_blocking_at;
		mov args[12], eax;
		pushad;
		push 15;
		call RunHookScript;
		popad;
		cmp cRet, 1;
		jl end;
		mov eax, rets[0];
end:
		hookend;
		retn;
	}
}
static const DWORD _roll_random=0x4A30C0;
static void __declspec(naked) ItemDamageHook() {
	__asm {
		hookbegin(6);
		mov args[0], eax; //min
		mov args[4], edx; //max
		mov args[8], edi; //weapon
		mov args[12], ecx; //critter
		mov args[16], esi; //type
		mov args[20], ebp; //non-zero for weapon melee attack
		test edi, edi;
		jnz skip;
		add args[16], 8;
skip:
		pushad;
		push 16;
		call RunHookScript;
		popad;
		cmp cRet, 0;
		je runrandom;
		mov eax, rets[0];
		cmp cRet, 1;
		je end;
		mov edx, rets[4];
runrandom:
		call _roll_random;
end:
		hookend;
		retn;
	}
}
DWORD _stdcall GetHSArgCount() {
	return ArgCount;
}
DWORD _stdcall GetHSArg() {
	if(cArg==ArgCount) return 0;
	else return args[cArg++];
}
void _stdcall SetHSArg(DWORD id, DWORD value) {
	if(id<ArgCount) args[id]=value;
}
void _stdcall GetHSArgs(char* data) {
	memcpy(data, args, ArgCount*4);
}
void _stdcall SetHSReturn(DWORD d) {
	if(cRet<8) rets[cRet++]=d;
}

void _stdcall RegisterHook(DWORD script, DWORD id) {
	if(id>=numHooks) return;
	for(DWORD i=0;i<hooks[id].size();i++) if(hooks[id][i]==script) return;
#ifdef TRACE
	char buf[256];
	sprintf_s(buf, "Global script %8x registered as hook id %d", script, id);
	dlogr(buf, DL_HOOK);
#endif
	hooks[id].push_back(script);
}
#define LoadHookScript(a,b) _LoadHookScript("data\\scripts\\hs_" a ".int", b)
static void _LoadHookScript(const char* path, int id) {
	WIN32_FIND_DATA file;
	HANDLE h;
	DWORD script;

	h=FindFirstFileA(path, &file);
	if(h!=INVALID_HANDLE_VALUE) {
		dlog("Loading hook script: ", DL_HOOK);
		dlogr(path, DL_HOOK);
		char* fName=file.cFileName;
		fName[strlen(fName)-4]=0;
		__asm {
			mov eax, fName;
			mov ebx, 0x004A3B74;
			call ebx;
			mov script, eax;
		}
		FindClose(h);
		if(script) hooks[id].push_back(script);
	}
}
static void HookScriptInit() {
	dlogr("Initing hook scripts", DL_HOOK|DL_INIT);

	LoadHookScript("tohit", 0);
	HookCall(0x421686, &ToHitHook);
	HookCall(0x4231D9, &ToHitHook);
	HookCall(0x42331F, &ToHitHook);
	HookCall(0x4237FC, &ToHitHook);
	HookCall(0x424379, &ToHitHook);
	HookCall(0x42438D, &ToHitHook);
	HookCall(0x42439C, &ToHitHook);
	HookCall(0x42679A, &ToHitHook);

	LoadHookScript("afterhitroll", 1);
	MakeCall(0x423893, &AfterHitRollHook, true);

	LoadHookScript("calcapcost", 2);
	HookCall(0x42307A, &CalcApCostHook);
	HookCall(0x42669F, &CalcApCostHook);
	HookCall(0x42687B, &CalcApCostHook);
	HookCall(0x42A625, &CalcApCostHook);
	HookCall(0x42A655, &CalcApCostHook);
	HookCall(0x42A686, &CalcApCostHook);
	HookCall(0x42AE32, &CalcApCostHook);
	HookCall(0x42AE71, &CalcApCostHook);
	HookCall(0x460048, &CalcApCostHook);
	HookCall(0x47807B, &CalcApCostHook);

	LoadHookScript("deathanim1", 3);
	LoadHookScript("deathanim2", 4);
	HookCall(0x4109DE, &CalcDeathAnimHook);

	LoadHookScript("combatdamage", 5);
	HookCall(0x423AB7, &CombatDamageHook);
	HookCall(0x423BBF, &CombatDamageHook);

	LoadHookScript("ondeath", 6);
	HookCall(0x4130CC, &OnDeathHook);
	HookCall(0x4130EF, &OnDeathHook);
	HookCall(0x413603, &OnDeathHook);
	HookCall(0x426EF0, &OnDeathHook);
	HookCall(0x42D1EC, &OnDeathHook);
	HookCall(0x42D6F9, &OnDeathHook);
	HookCall(0x457BC5, &OnDeathHook);
	HookCall(0x457E3A, &OnDeathHook);
	HookCall(0x457E54, &OnDeathHook);
	HookCall(0x4C14F9, &OnDeathHook);
	HookCall(0x425161, &OnDeathHook2);

	LoadHookScript("findtarget", 7);
	HookCall(0x429143, &FindTargetHook);

	LoadHookScript("useobjon", 8);
	HookCall(0x49C606, &UseObjOnHook);
	HookCall(0x473619, &UseObjOnHook);

	LoadHookScript("removeinvenobj", 9);
	MakeCall(0x477490, &RemoveObjHook, true);

	LoadHookScript("barterprice", 10);
	HookCall(0x474D4C, &BarterPriceHook);
	HookCall(0x475735, &BarterPriceHook);
	HookCall(0x475762, &BarterPriceHook);

	LoadHookScript("movecost", 11);
	HookCall(0x417665, &MoveCostHook);
	HookCall(0x44B88A, &MoveCostHook);

	LoadHookScript("hexmoveblocking", 12);
	LoadHookScript("hexaiblocking", 13);
	LoadHookScript("hexshootblocking", 14);
	LoadHookScript("hexsightblocking", 15);
	SafeWrite32(0x413979, (DWORD)&HexSightBlockingHook);
	SafeWrite32(0x4C1A88, (DWORD)&HexShootBlockingHook);
	SafeWrite32(0x423178, (DWORD)&HexShootBlockingHook);
	SafeWrite32(0x4232D4, (DWORD)&HexShootBlockingHook);
	SafeWrite32(0x423B4D, (DWORD)&HexShootBlockingHook);
	SafeWrite32(0x426CF8, (DWORD)&HexShootBlockingHook);
	SafeWrite32(0x42A570, (DWORD)&HexShootBlockingHook);
	SafeWrite32(0x42A0A4, (DWORD)&HexABlockingHook);
	MakeCall(0x48B848, &HexMBlockingHook, true);

	LoadHookScript("itemdamage", 16);
	HookCall(0x478560, &ItemDamageHook);

	dlogr("Completed hook script init", DL_HOOK|DL_INIT);

}

void HookScriptReload() {
	for(int i=0;i<numHooks;i++) hooks[i].clear();
	HookScriptInit();
	InitingHookScripts=1;
	for(int i=0;i<numHooks;i++) {
		if(hooks[i].size()) {
			DWORD script=hooks[i][0];
			__asm {
				mov edx, -1;
				mov ebx, 0x0046CCA4;
				mov eax, script;
				call ebx;
			}
		}
	}
	InitingHookScripts=0;
}