/*
 *    sfall
 *    Copyright (C) 2008, 2009, 2010, 2011, 2012  The sfall team
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <hash_map>

#include "main.h"
#include "ScriptExtender.h"
#include "HookScripts.h"
#include "input.h"
#include "LoadGameHook.h"
#include "version.h"
#include "vector9x.cpp"
#include "numbers.h"
#include "Logging.h"
#include "BarBoxes.h"
#include "Console.h"

#include "ScriptOps\ScriptArrays.hpp"
#include "ScriptOps\ScriptUtils.hpp"

#include "ScriptOps\WorldmapOps.hpp"
#include "ScriptOps\PerksOp.hpp"
#include "ScriptOps\MemoryOp.hpp"
#include "ScriptOps\StatsOp.hpp"
#include "ScriptOps\InterfaceOp.hpp"
#include "ScriptOps\GraphicsOp.hpp"
#include "ScriptOps\FileSystemOps.hpp"
#include "ScriptOps\ObjectsOps.hpp"
#include "ScriptOps\MiscOps.hpp"


typedef void (_stdcall *regOpcodeProc)(WORD opcode,void* ptr);

static DWORD highlightingToggled=0;
static DWORD MotionSensorMode;
static BYTE toggleHighlightsKey;
static int idle;
static char HighlightFail1[128];
static char HighlightFail2[128];

struct sGlobalScript {
	DWORD ptr;
	DWORD count;
	DWORD repeat;
	DWORD mode; //0 - local map loop, 1 - input loop, 2 - world map loop, 3 - local and world map loops

	sGlobalScript() {}
	sGlobalScript(DWORD script) {
		ptr=script;
		count=0;
		repeat=0;
		mode=0;
	}
};

static vector<DWORD> checkedScripts;
static vector<sGlobalScript> globalScripts;

stdext::hash_map<__int64, int> globalVars;
typedef stdext::hash_map<__int64, int> :: iterator glob_itr;
typedef stdext::hash_map<__int64, int> :: const_iterator glob_citr;
typedef std::pair<__int64, int> glob_pair;

stdext::hash_map<DWORD, sArrayVar> arrays;
typedef stdext::hash_map<DWORD, sArrayVar> :: iterator array_itr;
typedef stdext::hash_map<DWORD, sArrayVar> :: const_iterator array_citr;
typedef std::pair<DWORD, sArrayVar> array_pair;

DWORD nextarrayid=1;
vector<DWORD> tempArrays;

static void* opcodes[0x300];
DWORD AddUnarmedStatToGetYear=0;
DWORD AvailableGlobalScriptTypes=0;

static DWORD OverrideObj;
static DWORD OverrideScript;
struct {
	char bingle[0x3c];
	DWORD OverrideObj;
} OverrideStruct;

//eax contains the script pointer, edx contains the opcode*4

//To read a value, mov the script pointer to eax, call 0x4674F0, eax now contains the value type
//mov the script pointer to eax, call 0x467500, eax now contains the value

//To return a value, move it to edx, mov the script pointer to eax, call 0x4674DC
//mov the value type to edx, mov the script pointer to eax, call 0x46748C



static void _stdcall SetGlobalScriptRepeat2(DWORD script, DWORD frames) {
	for(DWORD d=0;d<globalScripts.size();d++) {
		if(globalScripts[d].ptr==script) {
			if(frames==0xFFFFFFFF) globalScripts[d].mode=!globalScripts[d].mode;
			else globalScripts[d].repeat=frames;
			break;
		}
	}
}
static void __declspec(naked) SetGlobalScriptRepeat() {
	__asm {
		push ebx;
		push ecx;
		push edx;
		mov ecx, eax;
		mov ebx, 0x4674F0;
		call ebx;
		mov edx, eax;
		mov eax, ecx;
		mov ebx, 0x467500;
		call ebx;
		cmp dx, 0xC001;
		jnz end;
		push eax;
		push ecx;
		call SetGlobalScriptRepeat2;
end:
		pop edx;
		pop ecx;
		pop ebx;
		retn;
	}
}
static void _stdcall SetGlobalScriptType2(DWORD script, DWORD type) {
	if(type>3) return;
	for(DWORD d=0;d<globalScripts.size();d++) {
		if(globalScripts[d].ptr==script) {
			globalScripts[d].mode=type;
			break;
		}
	}
}
static void __declspec(naked) SetGlobalScriptType() {
	__asm {
		push ebx;
		push ecx;
		push edx;
		mov ecx, eax;
		mov ebx, 0x4674F0;
		call ebx;
		mov edx, eax;
		mov eax, ecx;
		mov ebx, 0x467500;
		call ebx;
		cmp dx, 0xC001;
		jnz end;
		push eax;
		push ecx;
		call SetGlobalScriptType2;
end:
		pop edx;
		pop ecx;
		pop ebx;
		retn;
	}
}
static void __declspec(naked) GetGlobalScriptTypes() {
	__asm {
		push ebx;
		push ecx;
		push edx;
		mov edx, AvailableGlobalScriptTypes;
		mov ecx, eax;
		mov ebx, 0x4674DC;
		call ebx;
		mov edx, 0xc001;
		mov eax, ecx;
		mov ebx, 0x46748C;
		call ebx;
		pop edx;
		pop ecx;
		pop ebx;
		retn;
	}
}
static void SetGlobalVarInternal(__int64 var, int val) {
	glob_itr itr=globalVars.find(var);
	if(itr==globalVars.end()) globalVars.insert(glob_pair(var, val));
	else itr->second=val;
}
static void _stdcall SetGlobalVar2(const char* var, int val) {
	if(strlen(var)!=8) return;
	SetGlobalVarInternal(*(__int64*)var, val);
}
static void _stdcall SetGlobalVar2Int(DWORD var, int val) {
	SetGlobalVarInternal(var, val);
}
static void __declspec(naked) SetGlobalVar() {
	__asm {
		push ebx;
		push ecx;
		push edx;
		push edi;
		push esi;
		mov edi, eax;
		mov ebx, 0x4674F0;
		call ebx;
		mov eax, edi;
		mov ebx, 0x467500;
		call ebx;
		mov esi, eax;
		mov eax, edi;
		mov ebx, 0x4674F0;
		call ebx;
		mov edx, eax;
		mov eax, edi;
		mov ebx, 0x467500;
		call ebx;
		cmp dx, 0x9001;
		jz next;
		cmp dx, 0x9801;
		jz next;
		cmp dx, 0xc001;
		jnz end;
		push esi;
		push eax;
		call SetGlobalVar2Int;
		jmp end;
next:
		mov ebx, eax;
		mov eax, edi;
		mov ecx, 0x4678E0;
		call ecx;
		push esi;
		push eax;
		call SetGlobalVar2;
end:
		pop esi;
		pop edi;
		pop edx;
		pop ecx;
		pop ebx;
		retn;
	}
}
static DWORD GetGlobalVarInternal(__int64 val) {
	glob_citr itr=globalVars.find(val);
	if(itr==globalVars.end()) return 0;
	else return itr->second;
}
static DWORD _stdcall GetGlobalVar2(const char* var) {
	if(strlen(var)!=8) return 0;
	return GetGlobalVarInternal(*(__int64*)var);
}
static DWORD _stdcall GetGlobalVar2Int(DWORD var) {
	return GetGlobalVarInternal(var);
}
static void __declspec(naked) GetGlobalVarInt() {
	__asm {
		push ebx;
		push ecx;
		push edx;
		push edi;
		push esi;
		xor edx, edx;
		mov edi, eax;
		mov ebx, 0x4674F0;
		call ebx;
		mov esi, eax;
		mov eax, edi;
		mov ebx, 0x467500;
		call ebx;
		cmp si, 0x9001;
		jz next;
		cmp si, 0x9801;
		jz next;
		cmp si, 0xc001;
		jnz end;
		push eax;
		call GetGlobalVar2Int;
		mov edx, eax;
		jmp end;
next:
		mov edx, esi;
		mov ebx, eax;
		mov eax, edi;
		mov ecx, 0x4678E0;
		call ecx;
		push eax;
		call GetGlobalVar2;
		mov edx, eax;
end:
		mov eax, edi;
		mov ebx, 0x4674DC;
		call ebx;
		mov edx, 0xc001;
		mov eax, edi;
		mov ebx, 0x46748C;
		call ebx;
		pop esi;
		pop edi;
		pop edx;
		pop ecx;
		pop ebx;
		retn;
	}
}
static void __declspec(naked) GetGlobalVarFloat() {
	__asm {
		push ebx;
		push ecx;
		push edx;
		push edi;
		push esi;
		xor edx, edx;
		mov edi, eax;
		mov ebx, 0x4674F0;
		call ebx;
		mov esi, eax;
		mov eax, edi;
		mov ebx, 0x467500;
		call ebx;
		cmp si, 0x9001;
		jz next;
		cmp si, 0x9801;
		jz next;
		cmp si, 0xc001;
		jnz end;
		push eax;
		call GetGlobalVar2Int;
		mov edx, eax;
		jmp end;
next:
		mov edx, esi;
		mov ebx, eax;
		mov eax, edi;
		mov ecx, 0x4678E0;
		call ecx;
		push eax;
		call GetGlobalVar2;
		mov edx, eax;
end:
		mov eax, edi;
		mov ebx, 0x4674DC;
		call ebx;
		mov edx, 0xa001;
		mov eax, edi;
		mov ebx, 0x46748C;
		call ebx;
		pop esi;
		pop edi;
		pop edx;
		pop ecx;
		pop ebx;
		retn;
	}
}



static void __declspec(naked) GetSfallArg() {
	__asm {
		pushad;
		push eax;
		call GetHSArg;
		pop ecx;
		mov edx, eax;
		mov eax, ecx;
		call SetResult;
		mov eax, ecx;
		mov edx, 0xc001;
		call SetResultType;
		popad;
		retn;
	}
}
static DWORD _stdcall GetSfallArgs2() {
	DWORD args=GetHSArgCount();
	DWORD id=TempArray(args, 4);
	GetHSArgs(arrays[id].data);
	for(DWORD i=0;i<args;i++) arrays[id].types[i]=DATATYPE_INT;
	return id;
}
static void __declspec(naked) GetSfallArgs() {
	__asm {
		pushad;
		push eax;
		call GetSfallArgs2;
		pop ecx;
		mov edx, eax;
		mov eax, ecx;
		call SetResult;
		mov eax, ecx;
		mov edx, 0xc001;
		call SetResultType;
		popad;
		retn;
	}
}
static void __declspec(naked) SetSfallArg() {
	__asm {
		pushad;
		mov ecx, eax;
		call GetArgType;
		mov edi, eax;
		mov eax, ecx;
		call GetArg;
		mov edx, eax;
		mov eax, ecx;
		call GetArgType;
		mov esi, eax;
		mov eax, ecx;
		call GetArg;
		cmp di, 0xc001;
		jnz end;
		cmp si, 0xc001;
		jnz end;
		push edx;
		push eax;
		call SetHSArg;
end:
		popad;
		retn;
	}
}
static void __declspec(naked) SetSfallReturn() {
	__asm {
		push ebx;
		push ecx;
		push edx;
		mov ecx, eax;
		call GetArgType;
		mov edx, eax;
		mov eax, ecx;
		call GetArg;
		cmp dx, 0xc001;
		jnz end;
		push eax;
		call SetHSReturn;
end:
		pop edx;
		pop ecx;
		pop ebx;
		retn;
	}
}
static void __declspec(naked) InitHook() {
	__asm {
		push ecx;
		push edx;
		mov ecx, eax;
		mov edx, InitingHookScripts;
		call SetResult;
		mov eax, ecx;
		mov edx, 0xc001;
		call SetResultType;
		pop edx;
		pop ecx;
		retn;
	}
}


static void __declspec(naked) set_self() {
	__asm {
		pushad;
		mov ebp, eax;
		call GetArgType;
		mov edi, eax;
		mov eax, ebp;
		call GetArg;
		cmp di, 0xc001;
		jnz end;
		mov OverrideObj, eax;
		mov OverrideScript, ebp;
		mov OverrideStruct.OverrideObj, eax;
end:
		popad;
		retn;
	}
}
static void __declspec(naked) register_hook() {
	__asm {
		pushad;
		mov ebp, eax;
		call GetArgType;
		mov edi, eax;
		mov eax, ebp;
		call GetArg;
		cmp di, 0xc001;
		jnz end;
		push eax;
		push ebp;
		call RegisterHook;
end:
		popad;
		retn;
	}
}

static void __declspec(naked) sfall_ver_major() {
	__asm {
		push edx;
		push ecx;
		mov edx, VERSION_MAJOR;
		mov eax, ecx;
		call SetResult;
		mov eax, ecx;
		mov edx, 0xc001;
		call SetResultType;
		pop ecx;
		pop edx;
		retn;
	}
}
static void __declspec(naked) sfall_ver_minor() {
	__asm {
		push edx;
		push ecx;
		mov edx, VERSION_MINOR;
		mov eax, ecx;
		call SetResult;
		mov eax, ecx;
		mov edx, 0xc001;
		call SetResultType;
		pop ecx;
		pop edx;
		retn;
	}
}
static void __declspec(naked) sfall_ver_build() {
	__asm {
		push edx;
		push ecx;
		mov edx, VERSION_BUILD;
		mov eax, ecx;
		call SetResult;
		mov eax, ecx;
		mov edx, 0xc001;
		call SetResultType;
		pop ecx;
		pop edx;
		retn;
	}
}


//END OF SCRIPT FUNCTIONS
static const DWORD scr_find_sid_from_program=0x4A390C + 5;
static const DWORD scr_ptr=0x4A5E34 + 5;
static const DWORD scr_find_obj_from_program=0x4A39AC + 7;
static void __declspec(naked) FindSidHook() {
	__asm {
		cmp OverrideObj, 0;
		jz end;
		cmp eax, OverrideScript;
		jnz zero;
		lea eax, OverrideStruct;
		mov [edx], eax;
		mov eax, -2;
		mov OverrideObj, 0;
		mov OverrideScript, 0;
		retn;
zero:
		mov OverrideObj, 0;
		mov OverrideScript, 0;
end:
		push ebx;
		push ecx;
		push edx;
		push esi;
		push ebp;
		jmp scr_find_sid_from_program;
	}
}
static void __declspec(naked) ScrPtrHook() {
	__asm {
		cmp eax, -2;
		jnz end;
		xor eax, eax;
		retn;
end:
		push ebx;
		push ecx;
		push esi;
		push edi;
		push ebp;
		jmp scr_ptr;
	}
}
static void __declspec(naked) FindObjHook() {
	__asm {
		cmp OverrideObj, 0;
		jz end;
		cmp eax, OverrideScript;
		jnz zero;
		mov eax, OverrideObj;
		mov OverrideObj, 0;
		mov OverrideScript, 0;
		retn;
zero:
		mov OverrideObj, 0;
		mov OverrideScript, 0;
end:
		push ebx;
		push ecx;
		push edx;
		push esi;
		sub esp, 0xc;
		jmp scr_find_obj_from_program;
	}
}
static void __declspec(naked) MainGameLoopHook() {
	__asm {
		push ebx;
		push ecx;
		push edx;
		mov ebx, 0x004C8B78;
		call ebx;
		push eax;
		call RunGlobalScripts1;
		pop eax;
		pop edx;
		pop ecx;
		pop ebx;
		retn;
	}
}
static void __declspec(naked) CombatLoopHook() {
	__asm {
		pushad;
		call RunGlobalScripts1;
		popad;
		mov ebx, 0x004C8B78;
		jmp ebx;
	}
}
void ScriptExtenderSetup() {
#ifdef TRACE
	bool AllowUnsafeScripting=GetPrivateProfileIntA("Debug", "AllowUnsafeScripting", 0, ini)!=0;
#else
	const bool AllowUnsafeScripting=false;
#endif
	toggleHighlightsKey=GetPrivateProfileIntA("Input", "ToggleItemHighlightsKey", 0, ini);
	if(toggleHighlightsKey) MotionSensorMode=GetPrivateProfileIntA("Misc", "MotionScannerFlags", 1, ini);
	GetPrivateProfileStringA("Sfall", "HighlightFail1", "You aren't carrying a motion sensor.", HighlightFail1, 128, translationIni);
	GetPrivateProfileStringA("Sfall", "HighlightFail2", "Your motion sensor is out of charge.", HighlightFail2, 128, translationIni);

	idle=GetPrivateProfileIntA("Misc", "ProcessorIdle", -1, ini);
	modifiedIni=GetPrivateProfileIntA("Main", "ModifiedIni", 0, ini);

	dlogr("Adding additional opcodes", DL_SCRIPT);
	if(AllowUnsafeScripting) dlogr("  Unsafe opcodes enabled", DL_SCRIPT);
	else dlogr("  Unsafe opcodes disabled", DL_SCRIPT);

	SafeWrite32(0x46E370, 0x300);	//Maximum number of allowed opcodes
	SafeWrite32(0x46ce34, (DWORD)opcodes);	//cmp check to make sure opcode exists
	SafeWrite32(0x46ce6c, (DWORD)opcodes);	//call that actually jumps to the opcode
	SafeWrite32(0x46e390, (DWORD)opcodes);	//mov that writes to the opcode

	HookCall(0x480E7B, MainGameLoopHook); //hook the main game loop
	HookCall(0x422845, CombatLoopHook); //hook the combat loop

	MakeCall(0x4A390C, &FindSidHook, true);
	MakeCall(0x4A5E34, &ScrPtrHook, true);
	MakeCall(0x4A39AC, &FindObjHook, true);

	if(AllowUnsafeScripting) {
		opcodes[0x156]=ReadByte;
		opcodes[0x157]=ReadShort;
		opcodes[0x158]=ReadInt;
		opcodes[0x159]=ReadString;
	}
	opcodes[0x15a]=SetPCBaseStat;
	opcodes[0x15b]=SetPCExtraStat;
	opcodes[0x15c]=GetPCBaseStat;
	opcodes[0x15d]=GetPCExtraStat;
	opcodes[0x15e]=SetCritterBaseStat;
	opcodes[0x15f]=SetCritterExtraStat;
	opcodes[0x160]=GetCritterBaseStat;
	opcodes[0x161]=GetCritterExtraStat;
	opcodes[0x162]=funcTapKey;
	opcodes[0x163]=GetYear;
	opcodes[0x164]=GameLoaded;
	opcodes[0x165]=GraphicsFuncsAvailable;
	opcodes[0x166]=funcLoadShader;
	opcodes[0x167]=funcFreeShader;
	opcodes[0x168]=funcActivateShader;
	opcodes[0x169]=funcDeactivateShader;
	opcodes[0x16a]=SetGlobalScriptRepeat;
	opcodes[0x16b]=InputFuncsAvailable;
	opcodes[0x16c]=KeyPressed;
	opcodes[0x16d]=funcSetShaderInt;
	opcodes[0x16e]=funcSetShaderFloat;
	opcodes[0x16f]=funcSetShaderVector;
	opcodes[0x170]=funcInWorldMap;
	opcodes[0x171]=ForceEncounter;
	opcodes[0x172]=SetWorldMapPos;
	opcodes[0x173]=GetWorldMapXPos;
	opcodes[0x174]=GetWorldMapYPos;
	opcodes[0x175]=SetDMModel;
	opcodes[0x176]=SetDFModel;
	opcodes[0x177]=SetMoviePath;
	for(int i=0x178;i<0x189;i++)
		opcodes[i]=funcSetPerkValue;
	opcodes[0x189]=funcSetPerkName;
	opcodes[0x18a]=funcSetPerkDesc;
	opcodes[0x18b]=SetPipBoyAvailable;
	if(UsingExtraKillTypes()) {
		opcodes[0x18c]=GetKillCounter2;
		opcodes[0x18d]=ModKillCounter2;
	} else {
		opcodes[0x18c]=GetKillCounter;
		opcodes[0x18d]=ModKillCounter;
	}
	opcodes[0x18e]=GetPerkOwed;
	opcodes[0x18f]=SetPerkOwed;
	opcodes[0x190]=GetPerkAvailable;
	opcodes[0x191]=GetCritterAP;
	opcodes[0x192]=SetCritterAP;
	opcodes[0x193]=GetActiveHand;
	opcodes[0x194]=ToggleActiveHand;
	opcodes[0x195]=SetWeaponKnockback;
	opcodes[0x196]=SetTargetKnockback;
	opcodes[0x197]=SetAttackerKnockback;
	opcodes[0x198]=RemoveWeaponKnockback;
	opcodes[0x199]=RemoveTargetKnockback;
	opcodes[0x19a]=RemoveAttackerKnockback;
	opcodes[0x19b]=SetGlobalScriptType;
	opcodes[0x19c]=GetGlobalScriptTypes;
	opcodes[0x19d]=SetGlobalVar;
	opcodes[0x19e]=GetGlobalVarInt;
	opcodes[0x19f]=GetGlobalVarFloat;
	opcodes[0x1a0]=fSetPickpocketMax;
	opcodes[0x1a1]=fSetHitChanceMax;
	opcodes[0x1a2]=fSetSkillMax;
	opcodes[0x1a3]=EaxAvailable;
	//opcodes[0x1a4]=SetEaxEnvironmentFunc;
	opcodes[0x1a5]=IncNPCLevel;
	opcodes[0x1a6]=GetViewportX;
	opcodes[0x1a7]=GetViewportY;
	opcodes[0x1a8]=SetViewportX;
	opcodes[0x1a9]=SetViewportY;
	opcodes[0x1aa]=SetXpMod;
	opcodes[0x1ab]=SetPerkLevelMod;
	opcodes[0x1ac]=GetIniSetting;
	opcodes[0x1ad]=funcGetShaderVersion;
	opcodes[0x1ae]=funcSetShaderMode;
	opcodes[0x1af]=GetGameMode;
	opcodes[0x1b0]=funcForceGraphicsRefresh;
	opcodes[0x1b1]=funcGetShaderTexture;
	opcodes[0x1b2]=funcSetShaderTexture;
	opcodes[0x1b3]=funcGetTickCount;
	opcodes[0x1b4]=SetStatMax;
	opcodes[0x1b5]=SetStatMin;
	opcodes[0x1b6]=SetCarTown;
	opcodes[0x1b7]=fSetPCStatMax;
	opcodes[0x1b8]=fSetPCStatMin;
	opcodes[0x1b9]=fSetNPCStatMax;
	opcodes[0x1ba]=fSetNPCStatMin;
	opcodes[0x1bb]=fSetFakePerk;
	opcodes[0x1bc]=fSetFakeTrait;
	opcodes[0x1bd]=fSetSelectablePerk;
	opcodes[0x1be]=fSetPerkboxTitle;
	opcodes[0x1bf]=fIgnoreDefaultPerks;
	opcodes[0x1c0]=fRestoreDefaultPerks;
	opcodes[0x1c1]=fHasFakePerk;
	opcodes[0x1c2]=fHasFakeTrait;
	opcodes[0x1c3]=fAddPerkMode;
	opcodes[0x1c4]=fClearSelectablePerks;
	opcodes[0x1c5]=SetCritterHitChance;
	opcodes[0x1c6]=SetBaseHitChance;
	opcodes[0x1c7]=SetCritterSkillMod;
	opcodes[0x1c8]=SetBaseSkillMod;
	opcodes[0x1c9]=SetCritterPickpocket;
	opcodes[0x1ca]=SetBasePickpocket;
	opcodes[0x1cb]=SetPyromaniacMod;
	opcodes[0x1cc]=fApplyHeaveHoFix;
	opcodes[0x1cd]=SetSwiftLearnerMod;
	opcodes[0x1ce]=SetLevelHPMod;
	if(AllowUnsafeScripting) {
		opcodes[0x1cf]=WriteByte;
		opcodes[0x1d0]=WriteShort;
		opcodes[0x1d1]=WriteInt;
		for(int i=0x1d2;i<0x1dc;i++)
			opcodes[i]=CallOffset;
	}
	opcodes[0x1dc]=ShowIfaceTag;
	opcodes[0x1dd]=HideIfaceTag;
	opcodes[0x1de]=IsIfaceTagActive;
	opcodes[0x1df]=GetBodypartHitModifier;
	opcodes[0x1e0]=SetBodypartHitModifier;
	opcodes[0x1e1]=funcSetCriticalTable;
	opcodes[0x1e2]=funcGetCriticalTable;
	opcodes[0x1e3]=funcResetCriticalTable;
	opcodes[0x1e4]=GetSfallArg;
	opcodes[0x1e5]=SetSfallReturn;
	opcodes[0x1e6]=SetApAcBonus;
	opcodes[0x1e7]=GetApAcBonus;
	opcodes[0x1e8]=SetApAcEBonus;
	opcodes[0x1e9]=GetApAcEBonus;
	opcodes[0x1ea]=InitHook;
	opcodes[0x1eb]=GetIniString;
	opcodes[0x1ec]=funcSqrt;
	opcodes[0x1ed]=funcAbs;
	opcodes[0x1ee]=funcSin;
	opcodes[0x1ef]=funcCos;
	opcodes[0x1f0]=funcTan;
	opcodes[0x1f1]=funcATan;
	opcodes[0x1f2]=SetPalette;
	opcodes[0x1f3]=RemoveScript;
	opcodes[0x1f4]=SetScript;
	opcodes[0x1f5]=GetScript;
	opcodes[0x1f6]=NBCreateChar;
	opcodes[0x1f7]=fs_create;
	opcodes[0x1f8]=fs_copy;
	opcodes[0x1f9]=fs_find;
	opcodes[0x1fa]=fs_write_byte;
	opcodes[0x1fb]=fs_write_short;
	opcodes[0x1fc]=fs_write_int;
	opcodes[0x1fd]=fs_write_int;
	opcodes[0x1fe]=fs_write_string;
	opcodes[0x1ff]=fs_delete;
	opcodes[0x200]=fs_size;
	opcodes[0x201]=fs_pos;
	opcodes[0x202]=fs_seek;
	opcodes[0x203]=fs_resize;
	opcodes[0x204]=get_proto_data;
	opcodes[0x205]=set_proto_data;
	opcodes[0x206]=set_self;
	opcodes[0x207]=register_hook;
	opcodes[0x208]=fs_write_bstring;
	opcodes[0x209]=fs_read_byte;
	opcodes[0x20a]=fs_read_short;
	opcodes[0x20b]=fs_read_int;
	opcodes[0x20c]=fs_read_float;
	opcodes[0x20d]=list_begin;
	opcodes[0x20e]=list_next;
	opcodes[0x20f]=list_end;
	opcodes[0x210]=sfall_ver_major;
	opcodes[0x211]=sfall_ver_minor;
	opcodes[0x212]=sfall_ver_build;
	opcodes[0x213]=funcHeroSelectWin;
	opcodes[0x214]=funcSetHeroRace;
	opcodes[0x215]=funcSetHeroStyle;
	opcodes[0x216]=set_critter_burst_disable;
	opcodes[0x217]=get_weapon_ammo_pid;
	opcodes[0x218]=set_weapon_ammo_pid;
	opcodes[0x219]=get_weapon_ammo_count;
	opcodes[0x21a]=set_weapon_ammo_count;
	if(AllowUnsafeScripting) {
		opcodes[0x21b]=WriteString;
	}
	opcodes[0x21c]=get_mouse_x;
	opcodes[0x21d]=get_mouse_y;
	opcodes[0x21e]=get_mouse_buttons;
	opcodes[0x21f]=get_window_under_mouse;
	opcodes[0x220]=get_screen_width;
	opcodes[0x221]=get_screen_height;
	opcodes[0x222]=stop_game;
	opcodes[0x223]=resume_game;
	opcodes[0x224]=create_message_window;
	opcodes[0x225]=remove_trait;
	opcodes[0x226]=get_light_level;
	opcodes[0x227]=refresh_pc_art;
	opcodes[0x228]=get_attack_type;
	opcodes[0x229]=ForceEncounterWithFlags;
	opcodes[0x22a]=set_map_time_multi;
	opcodes[0x22b]=play_sfall_sound;
	opcodes[0x22c]=stop_sfall_sound;
	opcodes[0x22d]=create_array;
	opcodes[0x22e]=set_array;
	opcodes[0x22f]=get_array;
	opcodes[0x230]=free_array;
	opcodes[0x231]=len_array;
	opcodes[0x232]=resize_array;
	opcodes[0x233]=temp_array;
	opcodes[0x234]=fix_array;
	opcodes[0x235]=string_split;
	opcodes[0x236]=list_as_array;
	opcodes[0x237]=str_to_int;
	opcodes[0x238]=str_to_flt;
	opcodes[0x239]=scan_array;
	opcodes[0x23a]=get_tile_pid;
	opcodes[0x23b]=modified_ini;
	opcodes[0x23c]=GetSfallArgs;
	opcodes[0x23d]=SetSfallArg;
	opcodes[0x23e]=force_aimed_shots;
	opcodes[0x23f]=disable_aimed_shots;
	opcodes[0x240]=mark_movie_played;
	opcodes[0x241]=get_npc_level;
	opcodes[0x242]=set_critter_skill_points;
	opcodes[0x243]=get_critter_skill_points;
	opcodes[0x244]=set_available_skill_points;
	opcodes[0x245]=get_available_skill_points;
	opcodes[0x246]=mod_skill_points_per_level;
	opcodes[0x247]=set_perk_freq;
	opcodes[0x248]=get_last_attacker;
	opcodes[0x249]=get_last_target;
	opcodes[0x24a]=block_combat;
	opcodes[0x24b]=tile_under_cursor;
	opcodes[0x24c]=gdialog_get_barter_mod;
	opcodes[0x24d]=set_inven_ap_cost;
	opcodes[0x24e]=op_substr;
	opcodes[0x24f]=op_strlen;
	opcodes[0x250]=op_sprintf;
	// opcodes[0x251]=  RESERVED
	// opcodes[0x252]=  RESERVED
	opcodes[0x253]=op_typeof;
}

void LoadGlobalScripts() {
	HookScriptReload();
	dlogr("Loading global scripts", DL_SCRIPT);
	OverrideObj=0;
	OverrideScript=0;
	WIN32_FIND_DATA file;
	HANDLE h=FindFirstFileA("data\\scripts\\gl*.int", &file);
	if(h==INVALID_HANDLE_VALUE) return;
	char* fName=file.cFileName;
	DWORD script;
	DWORD found=1;
	while(found) {
		fName[strlen(fName)-4]=0;
		dlogr(fName, DL_SCRIPT);
		__asm {
			mov eax, fName;
			mov ebx, 0x004A3B74;
			call ebx;
			mov script, eax;
		}
		if(script) {
			sGlobalScript gscript=sGlobalScript(script);
			globalScripts.push_back(gscript);
			__asm {
				mov edx, -1;
				mov ebx, 0x0046CCA4;
				mov eax, script;
				call ebx;
			}
		}
		found=FindNextFileA(h, &file);
	}
	FindClose(h);
	dlogr("Finished loading global scripts", DL_SCRIPT);
	//ButtonsReload();
}

static DWORD _stdcall ScriptHasLoaded(DWORD script) {
	for(DWORD d=0;d<checkedScripts.size();d++) {
		if(checkedScripts[d]==script) return 0;
	}
	checkedScripts.push_back(script);
	return 1;
}

void ClearCheckedScripts() {
	checkedScripts.clear();
	globalScripts.clear(); //memory leak...

	//Reset some settable game values back to the defaults
	//Pyromaniac bonus
	SafeWrite8(0x424AB6, 5);
	//xp mod
	SafeWrite8(0x004AFAB8, 0x53);
	SafeWrite32(0x004AFAB9, 0x55575651);
	//Perk level mod
	SafeWrite32(0x00496880, 0x00019078);
	//HP bonus
	SafeWrite8(0x4AFBC1, 2);
	//Stat ranges
	StatsReset();
	//Bodypart hit chances
	*((DWORD*)0x510954)=GetPrivateProfileIntA("Misc", "BodypartHitMod0", 0xFFFFFFD8, ini);
	*((DWORD*)0x510958)=GetPrivateProfileIntA("Misc", "BodypartHitMod1", 0xFFFFFFE2, ini);
	*((DWORD*)0x51095C)=GetPrivateProfileIntA("Misc", "BodypartHitMod2", 0xFFFFFFE2, ini);
	*((DWORD*)0x510960)=GetPrivateProfileIntA("Misc", "BodypartHitMod3", 0x00000000, ini);
	*((DWORD*)0x510964)=GetPrivateProfileIntA("Misc", "BodypartHitMod4", 0xFFFFFFEC, ini);
	*((DWORD*)0x510968)=GetPrivateProfileIntA("Misc", "BodypartHitMod5", 0xFFFFFFEC, ini);
	*((DWORD*)0x51096C)=GetPrivateProfileIntA("Misc", "BodypartHitMod6", 0xFFFFFFC4, ini);
	*((DWORD*)0x510970)=GetPrivateProfileIntA("Misc", "BodypartHitMod7", 0xFFFFFFE2, ini);
	*((DWORD*)0x510974)=GetPrivateProfileIntA("Misc", "BodypartHitMod8", 0x00000000, ini);
	//skillpoints per level mod
	SafeWrite8(0x43C27a, 5);
}

static void RunScript(sGlobalScript* script) {
	script->count=0;
	DWORD sptr=script->ptr;
	__asm {
		mov edx, 1;
		mov eax, sptr;
		mov ebx, 0x0046DD2C;
		call ebx;
	}
}
static void RunGlobalScripts1() {
	if(idle>-1) Sleep(idle);
	if(toggleHighlightsKey) {
		//0x48C294 to toggle
		if(KeyDown(toggleHighlightsKey)) {
			if(!highlightingToggled) {
				if(MotionSensorMode&4) {
					DWORD scanner;
					static const DWORD _inven_pid_is_carried_ptr=0x471CA0;
					static const DWORD _item_m_dec_charges=0x4795A4;
					__asm {
						mov eax, ds:[0x6610B8];
						mov edx, 59;
						call _inven_pid_is_carried_ptr;
						mov scanner, eax;
					}
					if(scanner) {
						if(MotionSensorMode&2) {
							__asm {
								mov eax, scanner;
								call _item_m_dec_charges; //Returns -1 if the item has no charges
								inc eax;
								mov highlightingToggled, eax;
							}
							if(!highlightingToggled) DisplayConsoleMessage(HighlightFail2);
						} else highlightingToggled=1;
					} else {
						DisplayConsoleMessage(HighlightFail1);
					}
				} else highlightingToggled=1;
				if(highlightingToggled) {
					__asm {
						mov eax, dword ptr ds:[0x519578];
						xor ebx, ebx;
						mov edx, 0x10;
						mov ecx, 0x48C148;
						call ecx;
					}
				} else {
					highlightingToggled=2;
				}
			}
		} else if(highlightingToggled) {
			if(highlightingToggled==1) {
				__asm {
					mov eax, dword ptr ds:[0x519578];
					xor edx, edx;
					mov ecx, 0x48C1E4;
					call ecx;
				}
			}
			highlightingToggled=0;
		}
	}
	for(DWORD d=0;d<globalScripts.size();d++) {
		if(!globalScripts[d].repeat||(globalScripts[d].mode!=0&&globalScripts[d].mode!=3)) continue;
		if(++globalScripts[d].count>=globalScripts[d].repeat) {
			RunScript(&globalScripts[d]);
		}
	}
	if(tempArrays.size()) {
		for(DWORD i=0;i<tempArrays.size();i++) FreeArray(tempArrays[i]);
		tempArrays.clear();
	}
}
void RunGlobalScripts2() {
	if(idle>-1) Sleep(idle);
	for(DWORD d=0;d<globalScripts.size();d++) {
		if(!globalScripts[d].repeat||globalScripts[d].mode!=1) continue;
		if(++globalScripts[d].count>=globalScripts[d].repeat) {
			RunScript(&globalScripts[d]);
		}
	}
	if(tempArrays.size()) {
		for(DWORD i=0;i<tempArrays.size();i++) FreeArray(tempArrays[i]);
		tempArrays.clear();
	}
}
void RunGlobalScripts3() {
	if(idle>-1) Sleep(idle);
	for(DWORD d=0;d<globalScripts.size();d++) {
		if(!globalScripts[d].repeat||(globalScripts[d].mode!=2&&globalScripts[d].mode!=3)) continue;
		if(++globalScripts[d].count>=globalScripts[d].repeat) {
			RunScript(&globalScripts[d]);
		}
	}
	if(tempArrays.size()) {
		for(DWORD i=0;i<tempArrays.size();i++) FreeArray(tempArrays[i]);
		tempArrays.clear();
	}
}

void LoadGlobals(HANDLE h) {
	DWORD count, unused;
	ReadFile(h, &count, 4, &unused, 0);
	if(unused!=4) return;
	sGlobalVar var;
	for(DWORD i=0;i<count;i++) {
		ReadFile(h, &var, sizeof(sGlobalVar), &unused, 0);
		globalVars.insert(glob_pair(var.id, var.val));
	}
}
void LoadArrays(HANDLE h) {
	DWORD count, unused, id;
	ReadFile(h, &count, 4, &unused, 0);
	if(unused!=4) return;
	sArrayVar var;
	for(DWORD i=0;i<count;i++) {
		ReadFile(h, &id, 4, &unused, 0);
		ReadFile(h, &var, 8, &unused, 0);
		var.types=new DWORD[var.len];
		var.data=new char[var.len*var.datalen];
		ReadFile(h, var.types, 4*var.len, &unused, 0);
		ReadFile(h, var.data, var.len*var.datalen, &unused, 0);
		arrays.insert(array_pair(id, var));
	}
}
void SaveGlobals(HANDLE h) {
	DWORD count, unused;
	count=globalVars.size();
	WriteFile(h, &count, 4, &unused, 0);
	sGlobalVar var;
	glob_citr itr=globalVars.begin();
	while(itr!=globalVars.end()) {
		var.id=itr->first;
		var.val=itr->second;
		WriteFile(h, &var, sizeof(sGlobalVar), &unused, 0);
		itr++;
	}
}
void SaveArrays(HANDLE h) {
	DWORD count, unused, id;
	sArrayVar var;
	count=arrays.size();
	WriteFile(h, &count, 4, &unused, 0);
	array_citr itr=arrays.begin();
	while(itr!=arrays.end()) {
		id=itr->first;
		var=itr->second;
		WriteFile(h, &id, 4, &unused, 0);
		WriteFile(h, &var, 8, &unused, 0);
		WriteFile(h, var.types, 4*var.len, &unused, 0);
		WriteFile(h, var.data, var.len*var.datalen, &unused, 0);
		itr++;
	}
}
void ClearGlobals() {
	globalVars.clear();
	array_citr itr=arrays.begin();
	while(itr!=arrays.end()) {
		delete[] itr->second.data;
		delete[] itr->second.types;
		itr++;
	}
	arrays.clear();
}
int GetNumGlobals() { return globalVars.size(); }
void GetGlobals(sGlobalVar* globals) {
	glob_citr itr=globalVars.begin();
	int i=0;
	while(itr!=globalVars.end()) {
		globals[i].id=itr->first;
		globals[i++].val=itr->second;
		itr++;
	}
}
void SetGlobals(sGlobalVar* globals) {
	glob_itr itr=globalVars.begin();
	int i=0;
	while(itr!=globalVars.end()) {
		itr->second=globals[i++].val;
		itr++;
	}
}
int GetNumArrays() {
	return arrays.size();
}
void GetArrays(int* _arrays) {
	int pos=0;
	array_citr itr=arrays.begin();
	while(itr!=arrays.end()) {
		_arrays[pos++]=itr->first;
		_arrays[pos++]=itr->second.len;
		_arrays[pos++]=itr->second.datalen;
		itr++;
	}
}
void DEGetArray(int id, DWORD* types, void* data) {
	memcpy(types, arrays[id].types, arrays[id].len*4);
	memcpy(data, arrays[id].data, arrays[id].len*arrays[id].datalen);
}
void DESetArray(int id, const DWORD* types, const void* data) {
	if(types) memcpy(arrays[id].types, types, arrays[id].len*4);
	memcpy(arrays[id].data, data, arrays[id].len*arrays[id].datalen);
}
//fuctions to load and save appearance globals
void SetAppearanceGlobals(int race, int style) {
	SetGlobalVar2("HAp_Race", race);
	SetGlobalVar2("HApStyle", style);
}

void GetAppearanceGlobals(int *race, int *style) {
	*race=GetGlobalVar2("HAp_Race");
	*style=GetGlobalVar2("HApStyle");
}

