/*
 *    sfall
 *    Copyright (C) 2008, 2009, 2010, 2012  The sfall team
 *
 *    This program is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "main.h"
#include "scriptextender.h"
#include "vector9x.cpp"
#include <hash_map>
//#include "ScriptArrays.h"

extern stdext::hash_map<DWORD, sArrayVar> arrays;
extern DWORD nextarrayid;
extern vector<DWORD> tempArrays;


static DWORD _stdcall CreateArray(DWORD len, DWORD datasize) {
	if(datasize<4) datasize=4;
	if(datasize%4) datasize+=4-(datasize%4); //make sure each element is 4 byte aligned, and that there's always space for ints
	sArrayVar var;
	var.len=len;
	var.datalen=datasize;
	var.types=new DWORD[len];
	var.data=new char[len*datasize];
	memset(var.types, 0, len*4);
	memset(var.data, 0, len*datasize);
	while(arrays.find(nextarrayid)!=arrays.end()) nextarrayid++;
	arrays[nextarrayid]=var;
	return nextarrayid++;
}
static DWORD _stdcall TempArray(DWORD len, DWORD datasize) {
	DWORD id=CreateArray(len, datasize);
	tempArrays.push_back(id);
	return id;
}
static void _stdcall FreeArray(DWORD id) {
	if(arrays.find(id)==arrays.end()) return;
	delete[] arrays[id].types;
	delete[] arrays[id].data;
	arrays.erase(id);
}
static DWORD _stdcall GetArrayType(DWORD id, DWORD el) {
	if(arrays.find(id)==arrays.end() || arrays[id].len<=el) return 0;
	switch(arrays[id].types[el]) {
	case DATATYPE_NONE:
	case DATATYPE_INT:   return 0xc001;
	case DATATYPE_FLOAT: return 0xa001;
	case DATATYPE_STR:   return 0x9801;
	}
	return 0;
}
static DWORD _stdcall GetArrayData(DWORD id, DWORD el) {
	if(arrays.find(id)==arrays.end() || arrays[id].len<=el) return 0;
	switch(arrays[id].types[el]) {
	case DATATYPE_NONE:  return 0;
	case DATATYPE_INT:   
	case DATATYPE_FLOAT: return *(DWORD*)&arrays[id].data[el*arrays[id].datalen];
	case DATATYPE_STR:   return (DWORD)&arrays[id].data[el*arrays[id].datalen];
	}
	return 0;
}
static void _stdcall SetArray(DWORD id, DWORD el, DWORD datatype, DWORD val) {
	datatype&=0xffff;
	if(arrays.find(id)==arrays.end() || arrays[id].len<=el) return;
	switch(datatype) {
	case 0x9801:
	case 0x9001:
		arrays[id].types[el]=DATATYPE_STR;
		strcpy_s(&arrays[id].data[el*arrays[id].datalen], arrays[id].datalen, (char*)val);
		break;
	case 0xc001:
		arrays[id].types[el]=DATATYPE_INT;
		*(DWORD*)&arrays[id].data[el*arrays[id].datalen]=val;
		break;
	case 0xa001:
		arrays[id].types[el]=DATATYPE_FLOAT;
		*(DWORD*)&arrays[id].data[el*arrays[id].datalen]=val;
		break;
	}
}
static int _stdcall LenArray(DWORD id) {
	if(arrays.find(id)==arrays.end()) return -1;
	else return arrays[id].len;
}
static void _stdcall ResizeArray(DWORD id, DWORD newlen) {
	if(arrays.find(id)==arrays.end() || arrays[id].len==newlen) return;
	char* newdata=new char[newlen*arrays[id].datalen];
	DWORD* newtypes=new DWORD[newlen];
	if(newlen > arrays[id].len) {
		memcpy(newdata, arrays[id].data, arrays[id].len*arrays[id].datalen);
		memset(&newdata[arrays[id].len*arrays[id].datalen], 0, (newlen-arrays[id].len)*arrays[id].datalen);
		memcpy(newtypes, arrays[id].types, arrays[id].len*4);
		memset(&newtypes[arrays[id].len], 0, (newlen-arrays[id].len)*4);
	} else {
		memcpy(newdata, arrays[id].data, newlen*arrays[id].datalen);
		memcpy(newtypes, arrays[id].types, newlen*4);
	}
	delete[] arrays[id].data;
	delete[] arrays[id].types;
	arrays[id].data=newdata;
	arrays[id].types=newtypes;
	arrays[id].len=newlen;
}
static void _stdcall FixArray(DWORD id) {
	for(DWORD i=0;i<tempArrays.size();i++) {
		if(tempArrays[i]==id) tempArrays.remove_at(i--);
	}
}
static int _stdcall ScanArray(DWORD id, DWORD val, DWORD datatype) {
	datatype&=0xffff;
	if(datatype==0xc001) datatype=DATATYPE_INT;
	else if(datatype==0xa001) datatype=DATATYPE_FLOAT;
	else return -1;
	if(arrays.find(id)==arrays.end()) return -1;
	for(DWORD i=0;i<arrays[id].len;i++) {
		if(arrays[id].types[i]==datatype && *(DWORD*)&(arrays[id].data[i*arrays[id].datalen])==val) return i;
	}
	return -1;
}



static void __declspec(naked) create_array() {
	__asm {
		pushad;
		mov edi, eax;
		call GetArgType
		mov ebx, eax;
		mov eax, edi;
		call GetArg;
		mov ecx, eax;
		mov eax, edi;
		call GetArgType;
		mov edx, eax;
		mov eax, edi;
		call GetArg;
		cmp bx, 0xc001;
		jne fail;
		cmp dx, 0xc001;
		jne fail;
		push ecx;
		push eax;
		call CreateArray;
		mov edx, eax;
		jmp end;
fail:
		xor edx, edx;
end:
		mov eax, edi;
		call SetResult;
		mov edx, 0xc001;
		mov eax, edi;
		call SetResultType;
		popad;
		retn;
	}
}
static void __declspec(naked) set_array() {
	__asm {
		pushad;
		mov ebp, eax;
		//Get args
		call GetArgType
		mov edx, eax;
		mov eax, ebp;
		call GetArg;
		mov edi, eax;
		mov eax, ebp;
		call GetArgType
		mov ecx, eax;
		mov eax, ebp;
		call GetArg;
		mov esi, eax;
		mov eax, ebp;
		call GetArgType
		mov ebx, eax;
		mov eax, ebp;
		call GetArg;
		xchg eax, edi;
		//Error check
		cmp bx, 0xC001;
		jne end;
		cmp cx, 0xc001;
		jne end;
		cmp dx, 0x9001;
		jz next;
		cmp dx, 0x9801;
		jnz notstring;
next:
		mov ecx, edx;
		mov ebx, eax;
		mov eax, ebp;
		call GetStringVar;
		mov edx, ecx;
notstring:
		push eax;
		push edx;
		push esi;
		push edi;
		call SetArray
end:
		popad;
		retn;
	}
}
static void __declspec(naked) get_array() {
	__asm {
		pushad;
		mov ebp, eax;
		//Get args
		call GetArgType
		mov ebx, eax;
		mov eax, ebp;
		call GetArg;
		mov edi, eax;
		mov eax, ebp;
		call GetArgType
		mov ecx, eax;
		mov eax, ebp;
		call GetArg;
		mov esi, eax;
		mov eax, ebp;
		//Error check
		cmp bx, 0xc001;
		jne fail;
		cmp cx, 0xc001;
		jne fail;
		push edi;
		push esi;
		call GetArrayData;
		push eax;
		push edi;
		push esi;
		call GetArrayType;
		pop edx;
		mov ebx, eax;
		jmp end;
fail:
		xor edx, edx;
		mov ebx, 0xc001;
end:
		cmp bx, 0x9801;
		jne notstring;
		mov eax, ebp;
		call CreateStringVar;
		mov edx, eax;
notstring:
		mov eax, ebp;
		call SetResult;
		mov edx, ebx;
		mov eax, ebp;
		call SetResultType;
		popad;
		retn;
	}
}
static void __declspec(naked) free_array() {
	__asm {
		pushad;
		mov edi, eax;
		call GetArgType;
		mov ebx, eax;
		mov eax, edi;
		call GetArg;
		cmp bx, 0xc001;
		jne end;
		push eax;
		call FreeArray;
end:
		popad;
		retn;
	}
}
static void __declspec(naked) len_array() {
	__asm {
		pushad;
		mov edi, eax;
		call GetArgType;
		mov ebx, eax;
		mov eax, edi;
		call GetArg;
		cmp bx, 0xc001;
		jne fail;
		push eax;
		call LenArray;
		mov edx, eax;
		jmp end;
fail:
		xor edx, edx;
end:
		mov eax, edi;
		call SetResult;
		mov edx, 0xc001;
		mov eax, edi;
		call SetResultType;
		popad;
		retn;
	}
}
static void __declspec(naked) resize_array() {
	__asm {
		pushad;
		mov ebp, eax;
		//Get args
		call GetArgType
		mov ebx, eax;
		mov eax, ebp;
		call GetArg;
		mov edi, eax;
		mov eax, ebp;
		call GetArgType
		mov ecx, eax;
		mov eax, ebp;
		call GetArg;
		mov esi, eax;
		mov eax, ebp;
		//Error check
		cmp bx, 0xc001;
		jne end;
		cmp cx, 0xc001;
		jne end;
		push edi;
		push esi;
		call ResizeArray;
end:
		popad;
		retn;
	}
}
static void __declspec(naked) temp_array() {
	__asm {
		pushad;
		mov edi, eax;
		call GetArgType
		mov ebx, eax;
		mov eax, edi;
		call GetArg;
		mov ecx, eax;
		mov eax, edi;
		call GetArgType;
		mov edx, eax;
		mov eax, edi;
		call GetArg;
		cmp bx, 0xc001;
		jne fail;
		cmp dx, 0xc001;
		jne fail;
		push ecx;
		push eax;
		call TempArray;
		mov edx, eax;
		jmp end;
fail:
		xor edx, edx;
end:
		mov eax, edi;
		call SetResult;
		mov edx, 0xc001;
		mov eax, edi;
		call SetResultType;
		popad;
		retn;
	}
}
static void __declspec(naked) fix_array() {
	__asm {
		pushad;
		mov edi, eax;
		call GetArgType;
		mov ebx, eax;
		mov eax, edi;
		call GetArg;
		cmp bx, 0xc001;
		jne end;
		push eax;
		call FixArray;
end:
		popad;
		retn;
	}
}
static void __declspec(naked) scan_array() {
		__asm {
		pushad;
		mov ebp, eax;
		//Get args
		call GetArgType
		mov ebx, eax;
		mov eax, ebp;
		call GetArg;
		mov edi, eax;
		mov eax, ebp;
		call GetArgType
		mov ecx, eax;
		mov eax, ebp;
		call GetArg;
		mov esi, eax;
		mov eax, ebp;
		//Error check
		cmp bx, 0xc001;
		jne fail;
		cmp cx, 0xc001;
		je success;
		cmp cx, 0xa001;
		jne fail;
success:
		push ecx;
		push edi;
		push esi;
		call ScanArray;
		mov edx, eax;
		jmp end;
fail:
		xor edx, edx;
		dec edx;
end:
		mov ebx, 0xc001;
		mov eax, ebp;
		call SetResult;
		mov edx, ebx;
		mov eax, ebp;
		call SetResultType;
		popad;
		retn;
	}
}

// object LISTS

struct sList {
	DWORD* obj;
	DWORD len;
	DWORD pos;

	sList(const vector<DWORD>* vec) {
		len=vec->size();
		obj=new DWORD[len];
		for(DWORD i=0;i<len;i++) obj[i]=(*vec)[i];
		pos=0;
	}
};
static const DWORD obj_find_first_at_tile=0x48B5A8;
static const DWORD obj_find_next_at_tile=0x48B608;
static void FillListVector(DWORD type, vector<DWORD>& vec) {
	if(type==4) {
		DWORD** squares=(DWORD**)0x66BE08;
		for(int elv=0;elv<2;elv++) {
			DWORD* esquares=squares[elv];
			for(int tile=0;tile<10000;tile++) {
				esquares[tile]=0x8f000002;
			}
		}
	} else {
		for(int elv=0;elv<3;elv++) {
			for(int tile=0;tile<40000;tile++) {
				DWORD obj;
				__asm {
					mov edx, tile;
					mov eax, elv;
					call obj_find_first_at_tile;
					mov obj, eax;
				}
				while(obj) {
					DWORD otype = ((DWORD*)obj)[25];
					otype = (otype&0xff000000) >> 24;
					if(type==9 || (type==0&&otype==1) || (type==1&&otype==0) || (type>=2&&type<=5&&type==otype)) vec.push_back(obj);
					__asm {
						call obj_find_next_at_tile;
						mov obj, eax;
					}
				}
			}
		}
	}
}
static void* _stdcall list_begin2(DWORD type) {
	vector<DWORD> vec = vector<DWORD>();
	FillListVector(type, vec);
	sList* list=new sList(&vec);
	return list;
}
static DWORD _stdcall list_as_array2(DWORD type) {
	vector<DWORD> vec = vector<DWORD>();
	FillListVector(type, vec);
	DWORD id=TempArray(vec.size(), 4);
	for(DWORD i=0;i<vec.size();i++) {
		arrays[id].types[i]=DATATYPE_INT;
		((DWORD*)arrays[id].data)[i]=vec[i];
	}
	return id;
}
static DWORD _stdcall list_next2(sList* list) {
	if(list->pos==list->len) return 0;
	else return list->obj[list->pos++];
}
static void _stdcall list_end2(sList* list) {
	delete[] list->obj;
	delete list;
}
static void __declspec(naked) list_begin() {
	__asm {
		pushad;
		mov ebp, eax;
		call GetArgType;
		mov edi, eax;
		mov eax, ebp;
		call GetArg;
		cmp di, 0xc001;
		jnz fail;
		push eax;
		call list_begin2;
		mov edx, eax;
		jmp end;
fail:
		xor edx, edx;
		dec edx;
end:
		mov eax, ebp;
		call SetResult;
		mov eax, ebp;
		mov edx, 0xc001;
		call SetResultType;
		popad;
		retn;
	}
}
static void __declspec(naked) list_as_array() {
	__asm {
		pushad;
		mov ebp, eax;
		call GetArgType;
		mov edi, eax;
		mov eax, ebp;
		call GetArg;
		cmp di, 0xc001;
		jnz fail;
		push eax;
		call list_as_array2;
		mov edx, eax;
		jmp end;
fail:
		xor edx, edx;
		dec edx;
end:
		mov eax, ebp;
		call SetResult;
		mov eax, ebp;
		mov edx, 0xc001;
		call SetResultType;
		popad;
		retn;
	}
}
static void __declspec(naked) list_next() {
	__asm {
		pushad;
		mov ebp, eax;
		call GetArgType;
		mov edi, eax;
		mov eax, ebp;
		call GetArg;
		cmp di, 0xc001;
		jnz fail;
		push eax;
		call list_next2;
		mov edx, eax;
		jmp end;
fail:
		xor edx, edx;
		dec edx;
end:
		mov eax, ebp;
		call SetResult;
		mov eax, ebp;
		mov edx, 0xc001;
		call SetResultType;
		popad;
		retn;
	}
}
static void __declspec(naked) list_end() {
	__asm {
		pushad;
		mov ebp, eax;
		call GetArgType;
		mov edi, eax;
		mov eax, ebp;
		call GetArg;
		cmp di, 0xc001;
		jnz end;
		push eax;
		call list_end2;
end:
		popad;
		retn;
	}
}